@@ -26,9 +26,10 @@ import pwd
 import re
 
 import hsh
-import jobs
+import aliases
 import builtins
 import config
+import jobs
 
 # A command is an instruction issued to hsh by the user, typically causing the
 # creation of a subprocess executing the specified program.  It is represented
@@ -70,20 +71,35 @@ def format_cmd(args):
 # Command classes.
 
 class Token(object):
-    # Part of a command representing a syntactic token.
+    """Part of a command representing a syntactic token.
 
-    # This classes is a token superclass.  Subclasses need to implement these
-    # methods: __str__(), __len__(), modify().  Default implementation, useful
-    # to some subclasses, of these are provided: terminates(), terminated(),
-    # commences(). A version of modify() is provided which does a couple of
-    # useful checks and can be called with super().
+    The top of the token class hierarchy.  Subclasses need to implement methods
+    defined here.
 
-    # If dquotable is True, the the token may appear inside double quotes.
+    Additionally, subclasses may wish to override these class variables:
+
+    commencer_ch = None # Character indicating the start of the token
+    dquotable = False   # True to permit the token within double quotes.
+    """
+
+    commencer_ch = None
     dquotable = False
 
     def __init__(self):
         self.terminator = re.compile("()(\s)")
-        self.commencer = re.compile("\S")
+        if self.__class__.commencer_ch is not None:
+            self.commencer = re.compile(re.escape(self.__class__.commencer_ch))
+        else:
+            self.commencer = re.compile("\S")
+
+    def __str__(self):
+        """Return the raw text entered by the user.  There is no default
+        implementation of this method."""
+        raise Exception("Not Implemented")
+
+    def __len__(self):
+        "There is no default implementation of this method."
+        raise Exception("Not Implemented")
 
     def modify(self, txt="", begin=0, end=None):
         """Modify this token by replacing the text between begin and end with
@@ -97,6 +113,9 @@ class Token(object):
              from the next token.  Spill over is not possible in this case.
           3] boolean: True means that this token needs to be deleted.  May
              include spillover.
+
+        The default implementation checks the insertion range, raising an
+        exception in case of error, but does not perform the actual modify.
         """
         if end is None:
             end = self.__len__()
@@ -128,7 +147,13 @@ class Token(object):
         If possible completions exist, a list of strings is returned.  The
         first item is the root of the completions, the rest are possible
         extensions of the root.  Most tokens will return None if the cursor is
-        not at the end.  The default implementation returns None."""
+        not at the end.  The default implementation returns nothing."""
+        return []
+
+    def expand_directives(self):
+        """Return the list of directives which are expanded from this token.
+        The HashDirective token will return something here, but most other
+        tokens use the default implementation which returns an empty list."""
         return []
 
     def commences(self, txt):
@@ -166,7 +191,8 @@ class Token(object):
         return False
 
     def dump(self):
-        """Return a representation of the token tree as an array of strings."""
+        """Return a representation of the token tree as a list of strings.  The
+        default implementation returns an empty list."""
         return []
 
 class AggregateToken(Token):
@@ -390,6 +416,12 @@ class LeafToken(Token):
 class CommandText(AggregateToken):
     # The root of the token tree, contains a list of Args and Arg separators.
 
+    def __init__(self, do_aliases=True):
+        """Create a new CommandText token.  If aliases is False, do not expand
+        aliases in this command, to prevent unbounded recursion."""
+        super(CommandText, self).__init__()
+        self.do_aliases = do_aliases
+
     def insert(self, loc, txt):
         """Insert the provided text at the given location, raising an exception
         in case of error."""
@@ -404,16 +436,23 @@ class CommandText(AggregateToken):
         if re.match("\s", txt):
             return ArgSeparator()
         elif loc == 0:
-            return Prog()
+            return Prog(self.do_aliases)
+        elif txt[0] == "#":
+            return HashDirective()
         else:
             return Arg()
 
     def expand(self):
-        """Args are not concatenated on expansion and separators are
-        dropped since they return an empty list."""
+        """Command expansion is to concatenate the results of expansions of the
+        Args into a list of parameters.  ArgSeparators disappear since they
+        expand to the empty list."""
         res = []
         for tok in self.tokens:
             res.extend(tok.expand())
+        if (len(res) > 0) and not isinstance(self.tokens[0], Prog):
+            # No Prog token is present, so include an empty string so the first
+            # Arg won't be mistaken for the program.
+            res[0:0] = ('', )
         return res
 
     def completions(self, cloc):
@@ -424,14 +463,21 @@ class CommandText(AggregateToken):
             else:
                 cloc -= len(tok)
 
+    def expand_directives(self):
+        """Like parameter expansion, directives are concatenated into a list.
+        Directives which don't make sense together are filtered so that only
+        the last one is preserved."""
+        res = []
+        for tok in self.tokens:
+            res.extend(tok.expand_directives())
+        return HashDirective.canonify(res)
+
 class ArgOrProg(AggregateToken):
     # Intermediate class for implementation shared by Arg and Prog classes.
 
     def __init__(self):
         super(ArgOrProg, self).__init__()
-        esc_chars = "".join(map(lambda x: x.commencer_ch, arg_subtoks))
-        esc_chars += "*?[]"
-        self.escaper = re.compile("(["+ re.escape(esc_chars) +"])")
+        self.escaper = re.compile("(["+ re.escape(commencers + "*?[]") +"])")
 
     def create_token(self, txt, loc):
         if re.match("\s", txt):
@@ -501,109 +547,133 @@ class Arg(ArgOrProg):
         return text.split(os.sep)[-1:] + comps
 
 class Prog(ArgOrProg):
-    # Special case Arg specifying the program or builtin to be run.
+    """
+    Prog specifies what to run.  It must be the first token of a command.
+    
+    Prog objects behave differently depending on if they are specifying a
+    program, builtin or alias.  It would be desirable to use subclasses for the
+    three, but isn't straightforward because as text is entered the type would
+    change, resulting in a lot reparsing, and risk of infinite loops.
 
-    # Inherits basic mechanics of expansion from Arg, but has these additional
-    # restrictions:
+    The rules for choosing what to run are:
 
-    #  - It can only expand to a single value, otherwise an error results.
-    #  - Expansion is done from $PATH, not from the current directory.
-    #  - It may only exist as the very first token of a command.
+     - Supported syntax elements are expanded: single and double quoted tokens,
+       environment variables, named directories, job references.  Glob
+       expansion is not permitted in Progs.
 
-    # When executing programs there are security concerns to consider.  The
-    # user should not be "tricked" into running the wrong program.  Existing
-    # shells have restrictions in place to avoid this scenario.  The rules used
-    # by hsh are:
+     - If the resultant text starts with '+' then a builtin is being explicitly
+       chosen, and if there's no such builtin, then nothing is run.
 
-    #  - If the program name starts with '/', './' or '../' then the PATH will
-    #    not be consulted and normal file location rules will apply.  This
-    #    includes cases where tilde or environment expansion results in a
-    #    leading '/', './' or '../'.
-    #  - Otherwise, the program will be searched for relative to all entries on
-    #    the PATH, and in builtins.
+     - If the resultant text starts with '/', './' or '../' then an external
+       program is specified.  The program will be found absolutely or relative
+       to the current directory, and the PATH will not be consulted.
 
-    # This represents a departure from other shells (bash, zsh, ksh, tcsh) in
-    # the case of a command name with a '/' in it.  hsh will search for such
-    # commands in subdirectories of path entries, while other shells will run
-    # prgrams relative to the current directory.  With a well-structured path
-    # this does not represent a weakening of security, and in fact tighten
-    # security in same cases.  The old usage is still available by prepending
-    # "./" in that case.  Finally this allows namespacing of groups of
-    # commands, for instance, all of the ImageMagick programs could be moved to
-    # a subdirectory, so that their generic names would not be so confusing,
-    # and the foomatic programs could have simplified names in its own
-    # subdirectory.
+     - Alias list is searched, and if the text matches, that alias is expanded.
+       The alias text is expanded as a full command, except that the first
+       token will not do alias expansion.
 
-    # Glob expansion for progs is not supported.  Other shells do glob
-    # expansion from the current directory, so for instance 'ech? foo' will not
-    # call echo, but will instead look for a file in the current directory
-    # which matches the glob pattern.  hsh could search the path for progs
-    # matching the glob pattern, but that seems like a marginal feature.
+     - Builtin list is searched, and if the text matches, that builtin is used.
 
-    # re which matches commands not requiring PATH search
-    abs_cmd_re = re.compile(r"\.?\.?/")
+     - Finally, the text will be searched for relative to all entries on
+       the PATH.
 
-    def expand(self, check_exists=True):
-        """The command is expanded, with any glob characters causing an
-        exception.  \ escapes are removed.  If the command doesn't exist an
-        empty string is returned, unless check_exists is False."""
-        # The parent takes responsibility for all of the subtokens, which
-        # is done prior to glob expansion.
+    These rules depart from those of other shells (bash, zsh, ksh, tcsh) in the
+    case of a command name with a '/' in it.  hsh will search for such commands
+    in subdirectories of path entries, while other shells will run programs
+    relative to the current directory.  With a well-structured path this does
+    not represent a weakening of security, and in fact tightens security in
+    same cases.  The traditional usage is still available by prepending "./".
+    Finally this allows namespacing of groups of commands, for instance, all of
+    the ImageMagick programs could be moved to a subdirectory, so that their
+    generic names would not be so confusing, or the foomatic programs could
+    have simplified names and go in their own subdirectory.
+
+    As per shell tradition, external programs are searched only from the path
+    unless they start with './', '../', '/'.  This is so that the user isn't
+    tricked into running something in the local directory by mistake.
+
+    Glob expansion is not supported in Prog because multiple expansions just
+    opens risk of error, has non-obvious interaction with path search, and does
+    not seem particularly useful anyway.  Other syntax which may result in
+    multiple expansion may also be precluded.  This differs from other shells
+    which will look in the current directory when glob syntax is used.
+    """
+    # re which matches explicitly specified programs.
+    expl_cmd_re = re.compile(r"\.?\.?/")
+
+    def __init__(self, do_aliases):
+        super(Prog, self).__init__()
+        self.do_aliases = do_aliases
+
+    def _parexp(self):
+        """Return the result of the parent expanding the subtokens.  Remove
+        escapes because Prog does not do glob expansion.  Raise an exception if
+        the parent expands it to multiple tokens."""
         parexp = super(Prog, self).expand()
 
-        # An expression which expands to multiple tokens is not specific enough
-        # to be safe.
         if len(parexp) > 1:
             raise ParseException("Command not permitted as multiple expansion.")
 
         if len(parexp) == 0:
-            return []
+            return ""
 
         if glob_match.search(parexp[0]):
             raise ParseException("Glob symbols not permitted in command.")
 
-        expanded = [unescaper.sub("\\1", parexp[0])]
+        return unescaper.sub("\\1", parexp[0])
 
-        if Prog.abs_cmd_re.match(expanded[0]):
-            # Program is specified absolutely, so no need to use $PATH
-            if check_exists and not os.path.exists(expanded[0]):
-                return ['']
-        elif (expanded[0].startswith(config.builtin_commencer) or
-              expanded[0] in jobs.loaded_builtins.keys()):
-            # Builtin command
-            if expanded[0].startswith(config.builtin_commencer):
-                biname = expanded[0][len(config.builtin_commencer):]
+    def expand(self):
+        """The command is expanded, with any glob characters causing an
+        exception.  \ escapes are removed.  If the command doesn't exist an
+        empty string is returned."""
+
+        expnd = self._parexp()
+
+        if Prog.expl_cmd_re.match(expnd):
+            # Program is specified explicitly.
+            if os.path.exists(expnd):
+                return [expnd]
             else:
-                biname = expanded[0]
-            if biname in jobs.loaded_builtins.keys():
-                return [config.builtin_commencer + biname]
-            elif check_exists:
                 return ['']
+        elif expnd.startswith(config.builtin_commencer):
+            # Explicitly stated builtin
+            if (expnd[len(config.builtin_commencer):] in jobs.loaded_builtins):
+                return [expnd]
             else:
-                return expanded[0:1]
+                return ['']
+
+        elif expnd in aliases.manager and self.do_aliases:
+            return aliases.manager.get_parsed_value(expnd).expand()
+
+        elif expnd in jobs.loaded_builtins:
+            # Implicit builtin
+            return ['+' + expnd]
+
         else:
-            # Relative specification, so search the path.
-            expanded_abs = []
+            # Treat as a relative specification, so search the path.
+            expnd_mtch = []
 
             for pathe in os.environ["PATH"].split(os.path.pathsep):
-                ea = os.path.join(pathe, expanded[0])
-                if os.path.exists(ea) or not check_exists:
-                    expanded_abs.append(ea)
-                
-            if len(expanded_abs) == 0:
+                ea = os.path.join(pathe, expnd)
+                if os.path.exists(ea):
+                    expnd_mtch.append(ea)
+
+            if len(expnd_mtch) == 0:
                 return ['']
-            if check_exists and len(expanded_abs) > 1:
-                # Multiple commands with the same name appear on the path.
-                # The user should be informed of this irregularity.
-                expanded_abs = expanded_abs[:1]
+            if len(expnd_mtch) > 1:
+                hsh.display_obj.show_alert('Multiple commands found on path')
 
-            # And add a builtin too if needed.
-            if not check_exists or expanded[0] in jobs.loaded_builtins.keys():
-                expanded_abs[0:0] = [config.builtin_commencer + expanded[0]]
+            return expnd_mtch[0:1]
 
-            expanded = expanded_abs
+    def expand_directives(self):
+        """Only aliases can include directives."""
 
-        return expanded
+        expnd = self._parexp()
+
+        if expnd in aliases.manager and self.do_aliases:
+            return aliases.manager.get_parsed_value(expnd).expand_directives()
+        else:
+            return []
 
     def completions(self, cloc):
         """Completions are filename or built-in completions, and are searched
@@ -616,37 +686,90 @@ class Prog(ArgOrProg):
         if cloc != self.__len__(): # Cursor must be at the end
             return
 
-        # Fetch expansions without checking that they exist
-        expansions = self.expand(check_exists=False)
-        text = self.__str__()
+        expnd = self._parexp()
 
         comps = []
 
-        for exp in expansions:
-            # Use globbing to get external completions
-            if not exp.startswith(config.builtin_commencer):
-                ec = glob.glob(exp + '*')
-                # Append / to directories
-                ec = map(lambda x: os.path.isdir(x) and x+os.sep or x, ec)
-                # Trim root from completions
-                ec = map(lambda x: x[len(exp):], ec)
-                comps.extend(ec)
-                bimatch = exp
-            else:
-                # Trim the leading '+' from the text
-                bimatch = exp[len(config.builtin_commencer):]
+        if expnd.startswith(config.builtin_commencer):
+            # expand from the list of builtins
+            bimatch = expnd[len(config.builtin_commencer):]
             for biname in jobs.loaded_builtins.keys():
                 if biname.startswith(bimatch):
                     comps.append(biname[len(bimatch):])
 
+        else:
+            # Form a list of filenames on which to do glob expansion
+            if Prog.expl_cmd_re.match(expnd):
+                # Explicit specification so don't use $PATH
+                srchpaths = [expnd]
+            else:
+                srchpaths = map(lambda x: os.path.join(x, expnd),
+                                os.environ["PATH"].split(os.path.pathsep))
+
+            for sp in srchpaths:
+                gres = glob.glob(sp + "*")
+                # Append / to directories
+                gres = map(lambda x: os.path.isdir(x) and x+os.sep or x, gres)
+                # Trim the common bit off.
+                gres = map(lambda x: x[len(sp):], gres)
+                comps.extend(gres)
+
+            # Add any matching builtins
+            for biname in jobs.loaded_builtins.keys():
+                if biname.startswith(expnd):
+                    comps.append(biname[len(expnd):])
+
+            # Add any matching aliases
+            for alname in aliases.manager:
+                if alname.startswith(expnd) and self.do_aliases:
+                    comps.append(alname[len(expnd):])
+
+        # Sort and remove duplicates.
         comps.sort()
         i = len(comps) - 1
         while i > 0:
             if comps[i] == comps[i-1]:
                 del comps[i]
             i -= 1
+        # Escape syntax characters.
         comps = map(lambda x: self.escaper.sub(r"\\\1", x), comps)
-        return text.split(os.sep)[-1:] + comps
+        # Don't include directories in the common bit, to reduce clutter.
+        return self.__str__().split(os.sep)[-1:] + comps
+
+class HashDirective(LeafToken):
+    """A directive to the shell about how to launch the job, identified by a
+    leading # character."""
+
+    commencer_ch = "#"
+    # Grouped into mutually-exclusive sets.
+    directives_g = [["exclusive", "piped", "pty"],
+                    ["hide", "hideonsuccess", "show"]]
+    directives = reduce(lambda x,y: x+y, directives_g)
+
+    @staticmethod
+    def canonify(dlist):
+        """Given a list of directive names, remove incompatible ones and
+        duplicates, preferring the last of each group."""
+        for dg in HashDirective.directives_g:
+            dlist =  (filter(lambda x: x not in dg, dlist) +
+                      filter(lambda x: x in dg, dlist)[-1:])
+        return dlist
+
+    def expand(self):
+        """Hash directives are not included in the final command line."""
+        return []
+
+    def expand_directives(self):
+        return [self.__str__()[1:]]
+
+    def completions(self, cloc):
+        if cloc != self.__len__(): # Cursor must be at the end
+            return
+        root = self.__str__()[1:]
+        comps = filter(lambda x: x.startswith(root),HashDirective.directives)
+        comps = map(lambda x: x[len(root):] + " ", comps)
+        comps.sort()
+        return [root] + comps
 
 class ArgSeparator(LeafToken):
     # A token consisting of whitespace which separates Args.
@@ -664,8 +787,7 @@ class RawText(LeafToken):
     def __init__(self):
         super(RawText, self).__init__()
         # Terminates whenever any more specialized token starts.
-        term_chars = "".join(map(lambda x: x.commencer_ch, arg_subtoks))
-        self.terminator = re.compile("()(\s|["+ re.escape(term_chars) +"])")
+        self.terminator = re.compile("()(\s|["+ re.escape(commencers) +"])")
         self.commencer = re.compile("\S")
 
     def expand(self):
@@ -676,11 +798,6 @@ class SingleQuotedToken(LeafToken):
 
     commencer_ch = "'"
 
-    def __init__(self):
-        super(SingleQuotedToken, self).__init__()
-        # Uses custom terminate methods rather than the re
-        self.commencer = re.compile("'")
-
     def terminates(self, ch):
         # Terminates on a single quote other than the one at the beginning.
         if self.__len__() > 0 and ch[0:1] == "'":
@@ -722,9 +839,7 @@ class DoubleQuotedText(LeafToken):
     def __init__(self):
         super(DoubleQuotedText, self).__init__()
         # Terminates whenever any more specialized token starts.
-        term_chars = "".join(map(lambda x: x.commencer_ch,
-                                 filter(lambda x: x.dquotable, arg_subtoks)))
-        self.terminator = re.compile("()([\""+ re.escape(term_chars) +"])")
+        self.terminator = re.compile("()([\""+ re.escape(commencers_dq) +"])")
         self.commencer = re.compile('[^"]')
 
     def expand(self):
@@ -739,11 +854,6 @@ class DoubleQuotedToken(AggregateToken):
 
     commencer_ch = "\""
 
-    def __init__(self):
-        super(DoubleQuotedToken, self).__init__()
-        # Uses custom terminate methods rather than the re
-        self.commencer = re.compile('"')
-
     def terminates(self, ch):
         # Terminates on a double quote other than the one at the beginning.
         if self.__len__() > 0 and ch[0:1] == '"':
@@ -764,8 +874,8 @@ class DoubleQuotedToken(AggregateToken):
         elif txt.startswith('"'):
             return DoubleQuote()
 
-        for tt in arg_subtoks:
-            if tt.dquotable and txt.startswith(tt.commencer_ch):
+        for tt in arg_subtoks_dq:
+            if txt.startswith(tt.commencer_ch):
                 return tt()
 
         # No special token matches, so it's text
@@ -830,7 +940,6 @@ class EnvVariable(BracedToken):
     def __init__(self):
         super(EnvVariable, self).__init__()
         self.terminator = re.compile('()(\W)')
-        self.commencer = re.compile('\$')
 
     def expand(self):
         try:
@@ -851,7 +960,6 @@ class NamedDir(BracedToken):
     def __init__(self):
         super(NamedDir, self).__init__()
         self.terminator = re.compile('()(\W)')
-        self.commencer = re.compile('\~')
 
     def expand(self):
         return [escape_glob_chars(os.path.expanduser('~' + self._text()))]
@@ -870,7 +978,6 @@ class JobReference(BracedToken):
     def __init__(self):
         super(JobReference, self).__init__()
         self.terminator = re.compile('()(\s)')
-        self.commencer = re.compile('\%')
         self.jobid = None
         self.features = []
 
@@ -944,17 +1051,22 @@ class JobReference(BracedToken):
 
 ######################################################################
 
-# List of token classes which can be created within an arg.
+# Lists of token classes needed in parsing.
 
-# The list is used during parsing to decide which token class to create, and is
-# used during object construction to build terminator expressions.  Each class
-# in the list needs to specify two class variables:
-#   .commencer_ch: A character which identifies the beginning of such a token
-#   .dquotable: True to permit token in double quotes.  False by default
-
-# Superclasses which are not instantiated, such as ArgOrProg, and special
-# purpose tokens, such as DoubleQuote, are not included.
+# All token types which can be instantiated within an arg.  Excludes
+# superclasses which are not instantiated, like ArgOrProg, and special purpose
+# tokens, like DoubleQuote.  See Token class for explanation of class variables
+# commencer_ch, dquotable, etc.
 
 arg_subtoks = [ SingleQuotedToken, DoubleQuotedToken,
                 EnvVariable, NamedDir, JobReference,
               ]
+
+# A string consisting of all commencer characters.
+commencers = "".join(map(lambda x: x.commencer_ch, arg_subtoks))
+
+# Token types which can appear in double quotes.
+arg_subtoks_dq = filter(lambda x: x.dquotable, arg_subtoks)
+
+# A string consisting of all commencer characters which can appear in quotes
+commencers_dq = "".join(map(lambda x: x.commencer_ch, arg_subtoks_dq))
