@@ -29,7 +29,7 @@ builtin_commencer = '+'
 
 # Configuration and data directory access for hsh.
 
-version = '1.-2.1'
+version = '1.-2.2'
 
 conf_dir = os.path.join(os.environ["HOME"], ".hsh")
 jobs_dir = os.path.join(conf_dir, "jobs")
@@ -49,17 +49,22 @@ if not os.path.exists(jobs_dir):
 faces = {
     "default":      (curses.COLOR_WHITE,   curses.COLOR_BLACK),
     "alert":        (curses.COLOR_WHITE,   curses.COLOR_RED),
+    "wrapmark":     (curses.COLOR_CYAN,       curses.COLOR_BLACK),
+
     "input.header": (curses.COLOR_YELLOW, curses.COLOR_BLACK),
     "input.body":   (curses.COLOR_WHITE, curses.COLOR_BLACK),
     "job.header":   (curses.COLOR_WHITE, curses.COLOR_BLUE),
+    "job.headerwrap":  (curses.COLOR_CYAN,    curses.COLOR_BLUE),
     "job.stdout":   (curses.COLOR_WHITE,   curses.COLOR_BLACK),
     "job.stderr":   (curses.COLOR_RED,     curses.COLOR_BLACK),
     "job.stdin":    (curses.COLOR_YELLOW,  curses.COLOR_BLACK),
-    "job.in_buff":  (curses.COLOR_BLACK,   curses.COLOR_YELLOW),
+    "job.in_buff":     (curses.COLOR_WHITE, curses.COLOR_MAGENTA),
 
     "list.header":     (curses.COLOR_YELLOW,  curses.COLOR_BLACK),
     "list.highlight":  (curses.COLOR_WHITE,   curses.COLOR_BLUE),
     "list.highlight2": (curses.COLOR_YELLOW,  curses.COLOR_BLACK),
+
+    "kv.mod_buff":     (curses.COLOR_WHITE, curses.COLOR_MAGENTA),
 }
 
 # Output formats
@@ -79,34 +84,47 @@ header_job = ("%(username)s@%(hostname)s:%(cwd)s> %(cmd_line)s\n"
 listviews = [
 
     { 'name' : "SessionList",
-      'keybind' : curses.KEY_F3,
       'job_filter' : 
             lambda x: ((x.get_state() != "NotRun")
-                       and x.show_output()
                        and x.is_new),
       'header' : "    Session Jobs   \n",
       'display_format' : "%(username)s@%(hostname)s:%(cwd)s> %(expanded_cmd)s\n"
                   "(%(jobid)s) %(state)s - %(total_lines)s lines of output\n" },
 
     { 'name' : "RunningJobs",
-      'keybind' : curses.KEY_F4,
-      'job_filter' : lambda x: x.get_state() == "Running" and x.is_new,
+      'job_filter' : lambda x: x.get_state() in ["Running", "Initializing"]
+                               and x.is_new,
       'header' : "    Running Jobs   \n",
       'display_format' : "%(username)s@%(hostname)s:%(cwd)s> %(expanded_cmd)s\n"
                    "(%(jobid)s) %(state)s - %(total_lines)s lines of output\n"},
 
     { 'name' : "CommandHistory",
-      'keybind' : curses.KEY_F5,
       'job_filter' : lambda x: True,
       'highlight_face' : "list.highlight2",
       'display_format' : "%(jobid)s %(cmd_line)s" },
 
     { 'name' : "JobHistory",
-      'keybind' : curses.KEY_F6,
       'job_filter' : 
-            lambda x: ((x.get_state() != "NotRun") and x.show_output()),
+            lambda x: ((x.get_state() != "NotRun")),
       'header' : "    Full Job History   \n",
       'display_format' : "%(username)s@%(hostname)s:%(cwd)s> %(expanded_cmd)s\n"
                   "(%(jobid)s) %(state)s - %(total_lines)s lines of output\n" },
 
     ]
+
+# Typically useful default aliases
+
+default_aliases = {
+    'gpg' : 'gpg #exclusive',
+    'sudo' : 'sudo #exclusive',
+    'ssh' : 'ssh #exclusive',
+    'aptitude' : 'aptitude #exclusive',
+
+    'python' : 'python #pty',
+    'sh' : 'sh #pty',
+    'top' : 'top #pty',
+
+    'quit' : 'exit',
+    'logout' : 'exit',
+    'bye' : 'exit',
+}
