@@ -49,7 +49,11 @@ import errno
 import fcntl
 import logging
 import os
+import pty
+import signal
 import subprocess
+import sys
+import termios
 import time
 import thread
 import threading
@@ -58,6 +62,7 @@ import traceback
 import config
 import system
 import content
+import command
 
 # A mapping from command names to builtin classes.
 loaded_builtins = dict()
@@ -117,22 +122,24 @@ class JobManager(object):
         if listener in self.listeners:
             self.listeners.remove(listener)
 
-    def create_job(self, cmd, cmdline, start):
-        """Given a command as an array of args, create and return a new job to
-        execute it.  cmdline is the string entered by the user to generate the
-        command."""
+    def create_job(self, cmd, cmdline, directives, state):
+        """Create a new job from a command and parameters, assign it an id and
+        return it.
+        cmd: A list of args, the first being the job to execute.
+        cmdline: The string entered by the user, as a reference, but not used
+                 in job creation.
+        state: A string with the job's initial state
+        directives: A list of directives for how to start the job."""
         if cmd[0].startswith(config.builtin_commencer):
             biname = cmd[0][len(config.builtin_commencer):]
-            job = loaded_builtins[biname].create_job(cmd, cmdline, start)
+            jobcls = loaded_builtins[biname]
+            job = jobcls.create_job(cmd, cmdline, directives, state)
         else:
-            job = ExternalJob.create_job(cmd, cmdline, start)
+            job = ExternalJob.create_job(cmd, cmdline, directives, state)
         # Extend the all_jobs array to include the newly created job.
-        while len(self.all_jobs) < job.jid:
-            njdir = os.path.join(config.jobs_dir, str(len(self.all_jobs)))
-            if os.path.exists(njdir):
-                self.all_jobs.append(self.load_job(njdir))
-            else:
-                self.all_jobs.append(None)
+        for i in range(len(self.all_jobs), job.jid):
+            self.all_jobs.append(os.path.join(config.jobs_dir, str(i)))
+            self.get_job(i)
         self.all_jobs.append(job)
         for listener in self.listeners:
             listener.on_add_job(job)
@@ -163,7 +170,12 @@ class JobManager(object):
         if jobid >= len(self.all_jobs) or self.all_jobs[jobid] is None:
             return None
         if type(self.all_jobs[jobid]) == str:
-            self.all_jobs[jobid] = self.load_job(self.all_jobs[jobid])
+            try:
+                self.all_jobs[jobid] = self.load_job(self.all_jobs[jobid])
+            except:
+                logging.warn("Error loading job: %s" % jobid)
+                self.all_jobs[jobid] = None
+                return None
         return self.all_jobs[jobid]
 
     def get_prev_job(self, jobid):
@@ -210,9 +222,8 @@ class JobListener(object):
     """An "interface" for objects interested in changes to a job.  Particularly
     useful to an object displaying details of a job."""
 
-    def on_job_output(self, job, amount):
-        # Called when output has been added.  amount can be "line" or "char"
-        # indicating how much output has been appended.
+    def on_job_output(self, job):
+        # Called when output has been added to the job's content.
         pass
 
     def on_job_raw_output(self, job, text, channel):
@@ -246,14 +257,16 @@ class JobOutputThread(threading.Thread):
         self.outfile = file(os.path.join(self.job.jdir, channel), "w")
 
     def output_normal(self, text):
+        """Handle Job output in normal mode."""
         nr = self.job.output.append_region(text)
         nr.source = self.channel
         self.outfile.write(text)
         self.outfile.flush()
         for listener in self.job.listeners:
-            listener.on_job_output(self.job, "line")
+            listener.on_job_output(self.job)
 
     def output_raw(self, text):
+        """Handle Job output in raw mode."""
         for listener in self.job.listeners:
             listener.on_job_raw_output(self.job, text, self.channel)
 
@@ -270,11 +283,14 @@ class JobOutputThread(threading.Thread):
             while True:
                 if len(buf) < bufsize:
                     try:
-                        buf += self.pipe.read(bufsize)
-                    except IOError, ioe:
-                        # errno 11 - resource unavailable happens when no data
-                        # is ready.
-                        if ioe.errno != 11:
+                        if type(self.pipe) == file:
+                            buf += self.pipe.read(bufsize)
+                        elif type(self.pipe) == int:
+                            buf += os.read(self.pipe, bufsize)
+                    except Exception, e:
+                        # errno 11 - resource unavailable means no data is
+                        # ready, so keep going and try again.
+                        if not (type(e) in [IOError,OSError] and e.errno == 11):
                             raise
 
                 if len(buf) == 0:
@@ -284,11 +300,24 @@ class JobOutputThread(threading.Thread):
                     if ps_code is not None:
                         logging.info("Job terminated (%d %s)" %
                                      (ps_code, self.channel))
+                        self.job.ps_code = ps_code
                         break
                     else:
-                        ps_code = self.job.ps.poll()
-                        if ps_code is None:
-                            time.sleep(0.1)
+                        try:
+                            # XXX I'd like to use job.ps.poll() here, but it
+                            # fails, apparently because I changed the process
+                            # group earlier.
+                            (pid, ps_code) = os.waitpid(self.job.ps.pid,
+                                                        os.WNOHANG)
+                            if pid == 0:
+                                ps_code = None
+                                time.sleep(0.1)
+                        except OSError, ose:
+                            # The child may have gone away
+                            if ose.errno != errno.ECHILD:
+                                raise
+                            if not ps_code:
+                                ps_code = 0
 
                 elif raw_mode:
                     self.output_raw(buf)
@@ -314,38 +343,40 @@ class JobOutputThread(threading.Thread):
             logging.error(str(e))
             logging.error(traceback.format_exc())
 
-        logging.info("waiting for process")
-        self.job.ps_code = self.job.ps.wait()
         if self.job.ps_code == 0:
             self.job.state = "Succeeded"
         else:
             self.job.state = "Failed"
         self.job.pid = None
+        self.writeable = False
         self.job._write_job_state()
         for listener in self.job.listeners:
             listener.on_job_terminate(self.job)
         self.outfile.close()
 
 
-class Job():
+class Job(object):
 
     @classmethod
-    def create_job(cls, cmd, cmdline, start):
+    def create_job(cls, cmd, cmdline, directives, state="NotRun"):
         """Return a new job object created from the given command."""
         job = cls()
         job.cmd = cmd
         job.cmdline = cmdline
+        job.directives_in = directives
+        job._set_directives()
         job.cwd = os.getcwd()
         job.username = system.get_username()
         job.hostname = system.get_hostname()
         job.is_new = True
+        job.state = state
 
         # All jobs need an id, so get one for this job.  This is done by
         # finding the job directory with the highest number and adding one.
         # The os.mkdir() call returns an error if the directory exists, to
         # resolve collisions.
         while True:
-            job.jid = max([len(manager.all_jobs)]
+            job.jid = max([len(manager.all_jobs) - 1]
                           + map(int, os.listdir(config.jobs_dir))) + 1
             try:
                 job.jdir = os.path.join(config.jobs_dir, str(job.jid))
@@ -357,8 +388,6 @@ class Job():
 
         job._write_job_context()
         job._write_job_state()
-        if start:
-            job.start()
         return job
 
     @staticmethod
@@ -375,9 +404,10 @@ class Job():
             job = eval(clsname)()
         job.jdir = datadir
         job._load_params('creation_context',
-                         ("cmd", "cmdline",
-                          "cwd", "username", "hostname", "jid"))
-        job._load_params('job_state', ("state", "ps_code"))
+                         {"cmd":"", "cmdline":"", "directives_in":[],
+                          "cwd":"", "username":"", "hostname":"", "jid":""})
+        job._load_params('job_state', {"state":"", "ps_code":None})
+        job._set_directives()
         job.is_new = False
         return job
 
@@ -395,11 +425,31 @@ class Job():
         self.jid = None
         self.jdir = None
         self.listeners = list()
-        self.show_output_var = True
+        self.writeable = False
+
+        # For processes that use terminal output.
+        self.wants_terminal = False
+        self.has_terminal = False
+
+    def _set_directives(self):
+        """When directives_in and directives_default are set, use this to
+        set directives."""
+        if not hasattr(self,"directives_default"):
+            self.directives_default = []
+        dlist = self.directives_default + self.directives_in
+        self.directives = command.HashDirective.canonify(dlist)
+
+    def start(self):
+        if self.get_state() == "Running":
+            logging.info("Can't restart a running job")
+            return
+
+        self.state = "Running"
+        self.output[0:] = []
 
     def _write_job_context(self):
         self._write_params('creation_context',
-                           ("cmd", "cmdline",
+                           ("cmd", "cmdline", "directives_in",
                             "cwd", "username", "hostname", "jid"))
         clsfile = file(os.path.join(self.jdir, 'classname'), "w")
         clsfile.write(self.__class__.__name__)
@@ -418,13 +468,15 @@ class Job():
         cxtfile.close()
 
     def _load_params(self, name, attrs):
-        # Read job details from disk
+        # Read job details from disk.  name is the file to search, and attrs is
+        # a dict with keynames and default values for backward compatibility.
         ll = dict()
         execfile(os.path.join(self.jdir, name), dict(), ll)
-        for i in attrs:
+        for i in attrs.keys():
             try:
                 self.__dict__[i] = ll[name][i]
             except KeyError, ke:
+                self.__dict__[i] = attrs[i]
                 logging.info("Key missing from job info: %s, %s" % (name, i))
 
     def get_retcode(self):
@@ -436,8 +488,21 @@ class Job():
     def get_output(self):
         return self.output
 
-    def show_output(self):
-        return self.show_output_var
+    def can_send_input(self):
+        return False
+
+    def get_directives(self):
+        return self.directives
+
+    def send_input(self, inp):
+        """Send the provided text as input to the job.  Jobs which can accept
+        input will implement this."""
+        return
+
+    def close_input(self):
+        """Close the input stream to this job.  Job subclasses may implement
+        this."""
+        return
 
     def register_listener(self, listener):
         if not isinstance(listener, JobListener):
@@ -448,52 +513,218 @@ class Job():
     def unregister_listener(self, listener):
         if listener in self.listeners:
             self.listeners.remove(listener)
-    
+
+######################################################################
+# External jobs and some support methods.
+
+def new_pgrp():
+    """Support method which moves the current process into a new process group
+    and resets signal handlers."""
+    os.setpgid(os.getpid(), os.getpid())
+    signal.signal(signal.SIGINT, signal.SIG_DFL)
+    signal.signal(signal.SIGQUIT, signal.SIG_DFL)
+    signal.signal(signal.SIGTSTP, signal.SIG_DFL)
+    signal.signal(signal.SIGTTIN, signal.SIG_DFL)
+    signal.signal(signal.SIGTTOU, signal.SIG_DFL)
+    signal.signal(signal.SIGCHLD, signal.SIG_DFL)
+    return True
 
 class ExternalJob(Job):
     """A Job which executes an external command."""
 
     def start(self):
-        if self.get_state() == "Running":
-            logging.info("Can't retstart a running job")
+        super(ExternalJob, self).start()
+
+        self.ps_code = None
+        if "exclusive" in self.directives:
+            # An exclusive job is still remembered like other jobs, except its
+            # output and input are not recorded.  It is executed in the main
+            # thread, since not much else can go on while it's running.
+            self.wants_terminal = True
+            self.do_exclusive()
             return
 
-        self.state = "Running"
-        self.output[0:] = []
-        self.ps_code = None
-        self.ps = subprocess.Popen(self.cmd, cwd=self.cwd,
-                                   stdin=subprocess.PIPE,
-                                   stdout=subprocess.PIPE,
-                                   stderr=subprocess.PIPE)
-        self.pid = self.ps.pid
+        if "pty" in self.directives:
+            (mpty, spty) = pty.openpty()
+            self.stdin = mpty
+            self.stdout = mpty
+            self.ps = subprocess.Popen(self.cmd, cwd=self.cwd,
+                                       preexec_fn = lambda: new_pgrp(),
+                                       stdin=spty,
+                                       stdout=spty,
+                                       stderr=subprocess.PIPE)
+            self.stderr = self.ps.stderr
+        else:
+            self.ps = subprocess.Popen(self.cmd, cwd=self.cwd,
+                                       preexec_fn = (lambda: new_pgrp()),
+                                       stdin=subprocess.PIPE,
+                                       stdout=subprocess.PIPE,
+                                       stderr=subprocess.PIPE)
+            self.stdin = self.ps.stdin
+            self.stdout = self.ps.stdout
+            self.stderr = self.ps.stderr
 
-        self.stderrthread = JobOutputThread(self, self.ps.stderr, "stderr")
+        # To avoid race conditions, the parent process also puts the child in a
+        # new process group.
+        try:
+            os.setpgid(self.ps.pid, self.ps.pid)
+        except Exception, e:
+            logging.debug("failure setting pgid on child: " + str(e))
+
+        self.pid = self.ps.pid
+        self.writeable = True
+
+        self.stderrthread = JobOutputThread(self, self.stderr, "stderr")
         self.stderrthread.setDaemon(True)
         self.stderrthread.start()
 
-        self.stdoutthread = JobOutputThread(self, self.ps.stdout, "stdout")
+        self.stdoutthread = JobOutputThread(self, self.stdout, "stdout")
         self.stdoutthread.setDaemon(True)
         self.stdoutthread.start()
 
         self._write_job_state()
 
+    def do_exclusive(self, start_proc=True):
+        """Run the job in exclusive mode.  This function will block until the
+        job terminates, or its process is stopped.  The job will not be
+        permitted to run until the display thread grants control over the
+        terminal.  If start_proc is True, then the job process will be started,
+        otherwise it will be resumed with a CONT signal.
+
+        It's important that the job be correctly set up with the display, so
+        that it will be granted terminal access."""
+        if not self.wants_terminal:
+            logging.warn("do_exclusive(): self.wants_terminal must be true")
+            return
+
+        while not self.has_terminal:
+            time.sleep(0.1)
+
+        try:
+            if start_proc:
+                # Brand new job, so start the process.
+                self.tsettings =  None
+                self.ps = subprocess.Popen(self.cmd, cwd=self.cwd,
+                                           preexec_fn = lambda: new_pgrp())
+
+                # It would be best to put the child process into a new process
+                # group here also, but Popen waited for the preexec to finish,
+                # and setpgid fails with EACCES.
+                # os.setpgid(self.ps.pid, self.ps.pid)
+
+                # The child process has started already before it got into the
+                # foreground, and has probably STOPped because of it, so use
+                # the same code as for the resume case to put the job in the
+                # foreground and send it a CONT.  It would be better to put the
+                # process into the foreground in the preexec_fn, but that
+                # causes an exception internally within Python.
+
+                os.tcsetpgrp(sys.stdin.fileno(), os.getpgid(self.ps.pid))
+
+            elif not self.ps.pid:
+                logging.warn("Can't resume job, no pid")
+                return
+
+            # put process in foreground and send it the CONT signal
+            os.tcsetpgrp(sys.stdin.fileno(), os.getpgid(self.ps.pid))
+            if self.tsettings:
+                termios.tcsetattr(sys.stdin.fileno(), termios.TCSANOW,
+                                  self.tsettings)
+            os.kill(self.ps.pid, signal.SIGCONT)
+            self.state = "Running"
+
+        except OSError, ose:
+            if ose.errno == errno.ESRCH:
+                # The process is missing, so the job must have terminated.
+                # Just continue on and the situation will be resolved.
+                pass
+            else:
+                raise
+
+        # Wait for the process to terminate or STOP
+        try:
+            (pid, stat) = os.waitpid(self.ps.pid, os.WUNTRACED)
+        except OSError, ose:
+            if ose.errno == errno.ECHILD:
+                stat = 0
+            else:
+                raise
+
+        # Reclaim the foreground position.
+        os.tcsetpgrp(sys.stdin.fileno(), os.getpgrp())
+
+        if os.WIFSTOPPED(stat):
+            # The process is STOPped
+            logging.debug("Child process suspended")
+            self.tsettings = termios.tcgetattr(sys.stdin.fileno())
+            self.state = "Suspended"
+            self._write_job_state()
+        else:
+            # The process has terminated so clean up.
+            if not self.ps_code:
+                self.ps_code = stat
+            self.state = self.ps_code == 0 and "Succeeded" or "Failed"
+            self.wants_terminal = False
+            for listener in self.listeners:
+                listener.on_job_terminate(self)
+            self._write_job_state()
+
+    def suspend(self):
+        """Suspend the job.  Send a STOP signal to it.  This will only occur
+        from a different thread of control from that running the job and
+        communicates solely by sending a STOP signal to the job.  The main
+        thread takes responsibility for noticing that the job has stopped and
+        changing object state."""
+        if self.ps.pid and (self.get_state() == "Running"):
+            try:
+                os.kill(self.ps.pid, signal.SIGSTOP)
+            except OSError, ose:
+                if ose.errno != errno.ESRCH: # No such process
+                    raise
+
+    def can_send_input(self):
+        return self.writeable
+
+    def send_input(self, inp):
+        if not self.writeable:
+            raise Exception("Cannot write to job")
+        if type(self.stdin) == file:
+            self.stdin.write(inp)
+            self.stdin.flush()
+        elif type(self.stdin) == int:
+            while len(inp) > 0:
+                wc = os.write(self.stdin, inp)
+                inp = inp[wc:]
+        else:
+            raise Exception("Bad type for self.stdin: " + str(type(self.stdin)))
+
+    def close_input(self):
+        self.writeable = False
+        if type(self.stdin) == file:
+            self.stdin.close()
+        elif type(self.stdin) == int:
+            os.close(self.stdin)
+        else:
+            raise Exception("Bad type for self.stdin: " + str(type(self.stdin)))
+
 class BuiltInJob(Job):
     """A Job which executes a builtin command.  Actual builtins are classes
     which inherit from this one.  More complex ones override start() while
     simpler ones implement do_start()."""
 
+    def __init__(self):
+        super(BuiltInJob, self).__init__()
+        self.directives_default = ["hide"]
+
     def start(self):
-        if self.get_state() == "Running":
-            logging.info("Can't retstart a running job")
-            return
+        super(BuiltInJob, self).start()
 
-        self.state = "Running"
-        self.show_output_var = False
-        self.output[0:] = []
         try:
             self.do_start()
             self.state = "Succeeded"
             self._write_job_state()
+            for listener in self.listeners:
+                listener.on_job_output(self)
         except:
             self.state = "Failed"
             self._write_job_state()
